import struct
import sys

import wgpu
from enum import Enum, auto


class ErrorFlag(Enum):
    NORMAL = auto()
    NO_PIPELINE = auto()
    NO_BIND_GROUP = auto()
    NO_VERTEX_BUFFER = auto()
    NO_INDEX_BUFFER = auto()
    NULL_PIPELINE = auto()
    NULL_BIND_GROUP = auto()
    NULL_VERTEX_BUFFER = auto()
    NULL_INDEX_BUFFER = auto()
    NULL_INDIRECT_BUFFER = auto()
    WRONG_COLOR_ATTACHMENT = auto()
    MISSING_COLOR_ATTACHMENT = auto()


SHADER_SOURCE = """
    @group(0) @binding(0) var<uniform> offset: vec2f;

    @vertex
    fn vertex_main( @location(0) position: vec2f ) -> @builtin(position) vec4f {
        return vec4f(position + offset, 0, 1.0);
    }
    
    @fragment
    fn fragment_main() -> @location(0) vec4f { 
        return vec4f(); 
    }
"""

VERTEX_BUFFER_LAYOUT = {
    "array_stride": 4 * 2,
    "attributes": [
        {
            "format": wgpu.VertexFormat.float32x2,
            "offset": 0,
            "shader_location": 0,
        },
    ],
}


def run(use_render_bundle_encoder, error_flag):
    device = wgpu.utils.get_default_device()

    output_texture = device.create_texture(
        size=[128, 128],
        format=wgpu.TextureFormat.rgba8unorm,
        usage="RENDER_ATTACHMENT",
    )

    shader = device.create_shader_module(code=SHADER_SOURCE)

    pipeline = device.create_render_pipeline(
        layout="auto",
        vertex={
            "module": shader,
            "buffers": [VERTEX_BUFFER_LAYOUT],
        },
        fragment={
            "module": shader,
            "targets": [{"format": output_texture.format}],
        },
        primitive={"topology": "triangle-strip"},
    )

    data_buffer = device.create_buffer(size=2 * 4, usage="UNIFORM")

    vertex_buffer = device.create_buffer_with_data(
        data=struct.pack("8f", -0.5, -0.5, -0.5, +0.5, +0.5, -0.5, +0.5, +0.5),
        usage=wgpu.BufferUsage.VERTEX,
    )

    indirect_buffer = device.create_buffer_with_data(
        data=struct.pack("4i", 4, 1, 0, 0),
        usage=wgpu.BufferUsage.INDIRECT,
    )

    index_buffer = device.create_buffer_with_data(
        data=struct.pack("4i", 0, 1, 2, 3),
        usage=wgpu.BufferUsage.INDEX,
    )

    bind_group = device.create_bind_group(
        layout=pipeline.get_bind_group_layout(0),
        entries=[
            {"binding": 0, "resource": {"buffer": data_buffer}},
        ],
    )
    render_pass_descriptor = {
        "color_attachments": [
            {
                "clear_value": (0, 0, 0, 0),
                "load_op": wgpu.LoadOp.clear,
                "store_op": wgpu.StoreOp.store,
                "view": output_texture.create_view(),
            }
        ]
    }

    def do_draw(encoder):
        if error_flag != ErrorFlag.NO_PIPELINE:
            encoder.set_pipeline(pipeline)
        if error_flag != ErrorFlag.NO_BIND_GROUP:
            encoder.set_bind_group(0, bind_group)
        if error_flag != ErrorFlag.NO_VERTEX_BUFFER:
            encoder.set_vertex_buffer(0, vertex_buffer)
        if error_flag != ErrorFlag.NO_INDEX_BUFFER:
            encoder.set_index_buffer(index_buffer, "uint32")
        encoder.draw(4, 1, 0, 0)
        encoder.draw_indexed(4, 1, 0, 0)
        encoder.draw_indirect(indirect_buffer, 0)

    command_encoder = device.create_command_encoder()
    render_pass = command_encoder.begin_render_pass(**render_pass_descriptor)

    if use_render_bundle_encoder:
        color_formats = [output_texture.format]
        if error_flag == ErrorFlag.WRONG_COLOR_ATTACHMENT:
            color_formats = [wgpu.TextureFormat.rgba32float]
        elif error_flag == ErrorFlag.MISSING_COLOR_ATTACHMENT:
            color_formats = []
        render_bundle_encoder = device.create_render_bundle_encoder(
            color_formats=color_formats
        )
        do_draw(render_bundle_encoder)
        # Setting one or more of pipeline, bind_group, vertex_buffer, index_buffer
        # or indirect_buffer to None (and thereby having the GC release the object) before
        # the call to finish() will crash the program.
        if error_flag == ErrorFlag.NULL_PIPELINE:
            pipeline = None
        if error_flag == ErrorFlag.NULL_BIND_GROUP:
            bind_group = None
        if error_flag == ErrorFlag.NULL_VERTEX_BUFFER:
            vertex_buffer = None
        if error_flag == ErrorFlag.NULL_INDEX_BUFFER:
            index_buffer = None
        if error_flag == ErrorFlag.NULL_INDIRECT_BUFFER:
            indirect_buffer = None

        render_bundle = render_bundle_encoder.finish()
        # Setting the values to None **after** the call to finish() is fine.
        pipeline = bind_group = vertex_buffer = indirect_buffer = indexed_buffer = None
        render_pass.execute_bundles([render_bundle])
    else:
        do_draw(render_pass)
        # render_pass has no problem with them set to None before the call to end()
        pipeline = bind_group = vertex_buffer = indirect_buffer = indexed_buffer = None

    render_pass.end()
    device.queue.submit([command_encoder.finish()])


def main(argv):
    error_flag = None
    if len(sys.argv) == 2:
        try:
            error_flag = ErrorFlag[sys.argv[1].upper().replace("-", "_")]
        except KeyError:
            pass
    if error_flag:
        for use_render_bundle_encoder in (False, True):
            print("Normal" if not use_render_bundle_encoder else "RenderBundleEncoder")
            try:
                run(use_render_bundle_encoder, error_flag)
            except wgpu.GPUError as e:
                print(f'Caught exception "{type(e)}"')
    else:
        flags = [x.name.lower().replace("_", "-") for x in ErrorFlag]
        print("Usage: python RenderBundleEncoderBug.py <flag>")
        print("  where flag is one of ", ", ".join(flags))


if __name__ == "__main__":
    main(sys.argv)
